@@ -1,5 +1,5 @@
 /* This file is part of GNU Rush.                  
-   Copyright (C) 2008-2022 Sergey Poznyakoff
+   Copyright (C) 2008-2024 Sergey Poznyakoff
 
    GNU Rush is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -59,7 +59,7 @@ struct error_msg error_msg[] = {
 void
 set_error_msg(enum error_type type, char *text)
 {
-	error_msg[type].text = text;
+	error_msg[type].text = xstrdup(text);
 	error_msg[type].custom = 1;
 }
 
@@ -1139,6 +1139,7 @@ fork_process(struct rush_rule *rule, struct rush_request *req)
 	exit(0);
 }
 
+#if !HAVE_GETGROUPLIST
 static int
 membergid(gid_t gid, size_t gc, gid_t *gv)
 {
@@ -1148,18 +1149,44 @@ membergid(gid_t gid, size_t gc, gid_t *gv)
 			return 1;
 	return 0;
 }
+#endif
 
 static void
 get_user_groups(struct rush_request *req, size_t *pgidc, gid_t **pgidv)
 {
 	size_t gidc = 0, n = 0;
 	gid_t *gidv = NULL;
-	struct group *gr;
+	gid_t group = req->gid == NO_GID ? req->pw->pw_gid : req->gid;
 	
 	n = 32;
 	gidv = xcalloc(n, sizeof(gidv[0]));
 
-	gidv[0] = req->gid == NO_GID ? req->pw->pw_gid : req->gid;
+#if HAVE_GETGROUPLIST
+	int max_groups = (int) n;
+	for (;;) {
+		int ng = getgrouplist(req->pw->pw_name, group,
+				      gidv, &max_groups);
+		if (ng >= 0) {
+			/* On some systems getgrouplist returns 0 on success.
+			 */
+			gidc = ng > 0 ? ng : max_groups;
+			break;
+		}
+
+		/* Work around a bug on Darwin, where getgrouplist
+		   fails to increase max_groups. */
+		if (n == max_groups) {
+			if (INT_MAX / 3 * 2 <= n)
+				xalloc_die ();
+			max_groups += (max_groups + 1) / 2;
+		}
+		n = max_groups;
+		gidv = xrealloc(gidv, n * sizeof (gidv[0]));
+	}
+#else
+	struct group *gr;
+
+	gidv[0] = group;
 	gidc = 1;
 
 	setgrent();
@@ -1177,6 +1204,7 @@ get_user_groups(struct rush_request *req, size_t *pgidc, gid_t **pgidv)
 			}
 	}
 	endgrent();
+#endif
 	*pgidc = gidc;
 	*pgidv = gidv;
 }
